#######################################################################
# Database definitions
#######################################################################

# contains white-/blacklists
CREATE TABLE `greylisting_lists`
(
  `id` int unsigned NOT NULL auto_increment, # unique id
  `ipaddr` char(43),                         # IPv4/IPv6 IP-Address or CIDR notation in ascii
  `ipaddr_start` bigint unsigned NOT NULL,   # first IP-Address in range (in decimal, calculated by trigger)
  `ipaddr_end` bigint unsigned NOT NULL,     # last IP-Address in range (in decimal, calculated by trigger)
  `rcpt_to` varchar(255) default NULL,       # the recipient address
  `block_expires` datetime NOT NULL,         # the time that an initial block will/did expire
  `record_expires` datetime NOT NULL,        # the date after which this record is ignored
  `create_time` datetime NOT NULL,           # for timestamp of creation time of this record
  `last_update` timestamp NOT NULL default CURRENT_TIMESTAMP on update CURRENT_TIMESTAMP, # timestamp of last change to this record (automatic)
  `comment` varchar(255) NOT NULL,           # useable for comments
  PRIMARY KEY (`id`),
  KEY `ipaddr_start` (`ipaddr_start`),       # full index scan is faster then a full table scan
  KEY `ipaddr_end` (`ipaddr_end`),           #  -> http://kristiannielsen.livejournal.com/802.html
  KEY `rcpt_to` (`rcpt_to`(20))
);

# triggers for autocreation of ipaddr_start and ipaddr_end
DELIMITER ;;
/*!50003
CREATE TRIGGER `greylisting_lists_insert` BEFORE INSERT ON `greylisting_lists`
FOR EACH ROW
BEGIN
  SET @base = 32;
  IF !INSTR(NEW.`ipaddr`, '.') THEN
    SET @base = 128;
  END IF;
  IF NEW.`ipaddr` IS NULL THEN
    SET NEW.`ipaddr_start` = 0;
    SET NEW.`ipaddr_end` = 0;
  ELSEIF INSTR(NEW.`ipaddr`, '/') THEN
    SET NEW.`ipaddr_start` = inet_aton(substring_index(NEW.`ipaddr`, '/', 1));
    SET @ipaddr_count = POW(2, @base - substring_index(NEW.`ipaddr`, '/', -1));
    SET NEW.`ipaddr_end` = NEW.`ipaddr_start` + @ipaddr_count - 1;
  ELSE
    SET NEW.`ipaddr_start` = inet_aton(NEW.`ipaddr`);
    SET NEW.`ipaddr_end` = NEW.`ipaddr_start`;
  END IF;
END */;;

/*!50003
CREATE TRIGGER `greylisting_lists_update` BEFORE UPDATE ON `greylisting_lists`
FOR EACH ROW
BEGIN
  SET @base = 32;
  IF !INSTR(NEW.`ipaddr`, '.') THEN
    SET @base = 128;
  END IF;
  IF NEW.`ipaddr` IS NULL THEN
    SET NEW.`ipaddr_start` = 0;
    SET NEW.`ipaddr_end` = 0;
  ELSEIF INSTR(NEW.`ipaddr`, '/') THEN
    SET NEW.`ipaddr_start` = inet_aton(substring_index(NEW.`ipaddr`, '/', 1));
    SET @ipaddr_count = POW(2, @base - substring_index(NEW.`ipaddr`, '/', -1));
    SET NEW.`ipaddr_end` = NEW.`ipaddr_start` + @ipaddr_count - 1;
  ELSE
    SET NEW.`ipaddr_start` = inet_aton(NEW.`ipaddr`);
    SET NEW.`ipaddr_end` = NEW.`ipaddr_start`;
  END IF;
END */;;
DELIMITER ;

# contains data inserted during greylisting
CREATE TABLE `greylisting_data`
(
  `id` bigint unsigned NOT NULL auto_increment, # unique triplet id
  `relay_ip` char(39),                          # sending relay in IPv4/IPv6 ascii dotted quad notation
  `mail_from` varchar(255) default NULL,        # ascii address of sender
  `rcpt_to` varchar(255) default NULL,          # the recipient address.
  `block_expires` datetime NOT NULL,            # the time that an initial block will/did expire
  `record_expires` datetime NOT NULL,           # the date after which this record is ignored
  `blocked_count` bigint unsigned NOT NULL default '0', # num of blocked attempts to deliver
  `passed_count`  bigint unsigned NOT NULL default '0', # num of passed attempts we have allowed
  `aborted_count` bigint unsigned NOT NULL default '0', # num of attempts we have passed, but were later aborted
  `create_time` datetime NOT NULL,              # timestamp of creation time of this record
  `last_update` timestamp NOT NULL default CURRENT_TIMESTAMP on update CURRENT_TIMESTAMP, # timestamp of last change to this record (automatic)
  PRIMARY KEY (`id`),
  KEY `relay_ip` (`relay_ip`),
  KEY `mail_from` (`mail_from`(20)),            # To keep the index size down, only index first 20 chars
  KEY `rcpt_to` (`rcpt_to`(20))
);

# Example IP based wildcard whitelist + blacklist
#INSERT INTO `greylisting_lists` VALUES (0, "127.0.0.1", 0, 0, NULL,"0000-00-00 00:00:00", "9999-12-31 23:59:59", NOW(), NOW(), 'whitelist for 127.0.0.1');
#INSERT INTO `greylisting_lists` VALUES (0, "192.168.1.0/24", 0, 0, NULL, "9999-12-31 23:59:59", "9999-12-31 23:59:59", NOW(), NOW(), 'blacklist for 192.168.1.0/24');

# Example domain based wildcard whitelist + blacklist
#INSERT INTO `greylisting_lists` VALUES (0, NULL, 0, 0, "domain.com", "0000-00-00 00:00:00", "9999-12-31 23:59:59", NOW(), NOW(), 'whitelist for *@domain.com');
#INSERT INTO `greylisting_lists` VALUES (0, NULL, 0, 0, "sub.domain.com", "9999-12-31 23:59:59", "9999-12-31 23:59:59", NOW(), NOW(), 'blacklist for *@sub.domain.com');
